//
// Mork.cs: A parser for mork files (used by software such as Firefox and Thunderbird)
//
// Copyright (C) 2006 Pierre Östlund
//

//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Text.RegularExpressions;

namespace Beagle.Util
{
	public class MorkDatabase : IEnumerable {
		protected string mork_file;
		protected string enum_namespace;
		
		protected Hashtable dicts;
		protected Hashtable metadicts;
		protected Hashtable rows;
		protected Hashtable tables;
		
		protected string regex_dict = @"(?<id>\w+)\s*=(?<value>(?>[^()]+|\((?<n>)|\) (?<-n>))*)\)(\(|>)";
		protected string regex_row = @"(?<action>[-!+]?)\[(-|)(?<roid>[0-9A-Za-z:\^]+)(?<cells>(?>[^\[\]]+)?)\]";
		protected string regex_cell = @"\^(?<key>[0-9A-Fa-f]+)(\^(?<pvalue>[0-9A-Fa-f]+)|=(?<value>[0-9A-Fa-f]+))";
		protected string regex_table = @"{.*?:(?<ns>[0-9A-Fa-f\^]+) {\(k\^(?<tbl>[0-9A-Fa-f]+):c\)";

		public MorkDatabase (string mork_file)
		{
			this.mork_file = mork_file;
			this.dicts = new Hashtable ();
			this.metadicts = new Hashtable ();
			this.rows = new Hashtable ();
			this.tables = new Hashtable ();
		}
		
		public void Read ()
		{
			string content;
			StreamReader reader;
			
			reader = new StreamReader (mork_file);
			content  = reader.ReadToEnd ();
			reader.Close ();
			
			Reset ();
			Read (content);
		}

		protected void Read (string content)
		{
			int position = -1;

			while (++position != content.Length) {
				
				if (content [position].Equals ('/') && content [position].Equals ('/'))
					// Ignore comments
					position = content.IndexOf ('\n');
				else if (content [position].Equals ('<') && content [position+2].Equals ('<'))
					// Parse metadict information
					ParseDict (Read (content, ref position, "<(", ")>"), true);
				else if (content [position].Equals ('<'))
					// Parse dict information
					ParseDict (Read (content, ref position, "<(", ")>"), false);
				else if (content [position].Equals ('{')) {
					// Parse table information
					ParseTable (Read (content, ref position, "{", "}"));
				 }else if (content [position].Equals ('[')) 
					// Parse rows
					ParseRows (Read (content, ref position, "[", "]"), null, null);
				else if (content [position].Equals ('@') && content [position+1].Equals ('$'))
					// Parse groups
					ParseGroups (Read (content, ref position, "@$${", "@$$}"));
			}
		}
		
		protected string Read (string content, ref int position, string start, string end)
		{
			int tmp = position, start_position = position;
			
			do {
				position = content.IndexOf (end, position+1);
				if ((tmp = content.IndexOf (start, tmp+1)) < 0)
					break;
			} while (tmp < position);
			
			return content.Substring (start_position, position-start_position+1);
		}
		
		protected virtual void ParseDict (string dict, bool metadict)
		{
			Regex reg = new Regex (@"(?<id>[0-9A-Fa-f]+)\s*=(?<value>(.*))", RegexOptions.Compiled);
			
			// This is sooo lame that, but it's an easy solution that works
			foreach (string t in Regex.Replace (dict.Substring (2, dict.Length-3).Replace ("\n", ""), @"\)\s*\(", "\n").Split ('\n')) {
				
				Match m = reg.Match (t);
				
				if (metadict) 
					metadicts [m.Result ("${id}")] = m.Result ("${value}");
				else
					dicts [m.Result ("${id}")] = Decode (m.Result ("${value}"), "iso-8859-1");;
			}
			
		}
		
		protected virtual void ParseTable (string table)
		{
			int start = table.IndexOf ('}')+1;
			Match m = new Regex (regex_table, RegexOptions.Compiled).Match (table);
			
			ParseRows (table.Substring (start, table.Length-start-1), m.Result ("${ns}"), m.Result ("${tbl}"));
		}
		
		protected virtual void ParseRows (string rows, string ns, string table)
		{
			Regex reg = new Regex (regex_row, RegexOptions.Compiled);
			
			foreach (Match m in reg.Matches (Clean (rows))) {
				// tmp [0] == id, tmp [1] == ns
				string[] tmp = m.Result ("${roid}").Split (':');
				
				if (m.Result ("${action}") == "-" || m.Result ("${cells}") == string.Empty)
					RemoveRow (tmp [0], (tmp.Length > 1 ? tmp [1] : ns));
				else
					AddRow (tmp [0], (tmp.Length > 1 ? tmp [1] : ns), table, m.Result ("${cells}"));
			}
		}
		
		protected virtual void ParseGroups (string groups)
		{
			int start = groups.IndexOf ("{@")+2;
			
			Read (groups.Substring (start, groups.Length-start-1));
		}
		
		protected string Clean (string str)
		{
			return str.Replace ("\n", "").Replace (" ", "");
		}
		
		public string ParseNamespace (string ns)
		{
			if (ns.StartsWith ("^"))
				return ns;
			else {
				foreach (string key in metadicts.Keys)
					if ((metadicts [key] as string) == ns)
						return String.Format ("^{0}", key);
			}
			
			return ns;
		}
		
		public void AddRow (string id, string ns, string table, string cols)
		{
			string ns2 = ParseNamespace (ns);
			
			if (id == string.Empty || ns2 == string.Empty || cols == string.Empty)
				return;
			else if (!rows.ContainsKey (ns2))
				rows [ns2] = new Hashtable ();
			
			// Add row with a twist... (aka hack)
			if (!Exists (id, ns2))
				(rows [ns2] as Hashtable).Add (id, cols);
			else if (cols.Length >= GetCells (id, ns2).Length)
				(rows [ns2] as Hashtable) [id] = cols;
			else
				(rows [ns2] as Hashtable) [id] = String.Format ("{0}{1}", (rows [ns2] as Hashtable) [id], cols);
			
			// Save table association
			if (!tables.ContainsKey (id))
				tables [id] = table;
		}
		
		public void RemoveRow (string id, string ns)
		{
			string ns2 = ParseNamespace (ns);
			
			if (!rows.ContainsKey (ns2))
				return;
			
			(rows [ns2] as Hashtable).Remove (id);
			tables.Remove (id);
		}
		
		public string GetCells (string id, string ns)
		{
			string ns2 = ParseNamespace (ns);
			
			return (ns2 != null ?(rows [ns2] as Hashtable) [id] as string : null);
		}
		
		public Hashtable Compile (string id, string ns)
		{
			string ns2 = ParseNamespace (ns);
			
			if (!Exists (id, ns2))
				return null;
			
			Hashtable tbl = new Hashtable ();
			Regex reg = new Regex (regex_cell, RegexOptions.Compiled);
			
			foreach (Match m in reg.Matches (GetCells (id, ns2))) {
				tbl [metadicts [m.Result ("${key}")]] = (m.Result ("${pvalue}") != string.Empty ? 
					dicts [m.Result("${pvalue}")] : m.Result ("${value}"));
			}
			
			tbl ["id"] = id;
			tbl ["table"] = tables [id];
			
			return tbl;
		}

		public bool Exists (string id, string ns)
		{
			string ns2 = ParseNamespace (ns);
			
			return (ns2 != null ? (rows [ns] as Hashtable).ContainsKey (id) : false);
		}
		
		public int GetRowCount (string ns)
		{
			string ns2 = ParseNamespace (ns);
			
			if (ns2 == null || rows [ns2] == null)
				return -1;
			
			return (rows [ns2] as Hashtable).Count;
		}
		
		public int GetRowCount (string ns, string table)
		{
			int count = 0;
			string ns2 = ParseNamespace (ns);
			
			if (ns2 == null || rows [ns2] == null)
				return -1;
			
			foreach (string id in (rows [ns2] as Hashtable).Keys) {
				if ((string) tables [id] == table)
					count++;
			}
			
			return count;
		}
		
		public IEnumerator GetEnumerator ()
		{
			string ns = ParseNamespace (EnumNamespace);
			
			if (ns == null || (rows [ns] as Hashtable) == null || Empty)
				return null;
			
			return (rows [ns] as Hashtable).Keys.GetEnumerator ();
		}
		
		public void Reset ()
		{
			dicts.Clear ();
			metadicts.Clear ();
			rows.Clear ();
			tables.Clear ();
		}
		
		// A _very_ hackish method to convert an UTF-8-morkish-encoded string to another encoding
		// (see Encoding-property)
		public static string Decode (string str, string encoding)
		{
			int start = 0, old;
			string decoded = string.Empty;
			
			if (str == string.Empty || encoding == string.Empty || str == null || encoding == null)
				return str;
			
			foreach (Match m in Regex.Matches (str, @"\$(?<1>[0-9A-Fa-f]{2})\$(?<2>[0-9A-Fa-f]{2})")) {
				old = start;
				start = m.Index+6;
				
				byte[] b = new byte [] { 
					Convert.ToByte (Thunderbird.Hex2Dec (m.Result ("${1}"))), 
					Convert.ToByte (Thunderbird.Hex2Dec (m.Result ("${2}"))) };
				
				byte[] tmp = System.Text.Encoding.Convert (
					System.Text.Encoding.UTF8, 
					System.Text.Encoding.GetEncoding (encoding), b);
				
				decoded = String.Concat (decoded, 
					str.Substring (old, start-old-6), 
					System.Text.Encoding.GetEncoding (encoding).GetString (tmp));
			}
			
			if (start == 0 && decoded.Length == 0)
				return str;
			else if (start != decoded.Length)
				decoded = String.Concat (decoded, str.Substring (start, str.Length-start));
			
			return decoded;
		}
		
		public int Rows {
			get {
				int count = 0;
				
				foreach (Hashtable r in rows.Values)
					count += r.Count;
			
				return count;
			}
		}
		
		public string EnumNamespace {
			get { return enum_namespace; }
			set { enum_namespace = value; }
		}
		
		public string Filename {
			get { return mork_file; }
		}

		// There will always exist an item with id 1 in namespace 80, which means
		// that when there are less than two items in the database, it's empty
		public bool Empty {
			get { return (rows.Count > 1 ? false : true); }
		}
	}
}
